@@ -1,4 +1,5 @@
 import asyncio
+import copy
 import os
 import inspect
 import json
@@ -19,14 +20,16 @@ import aiohttp.web as aiohttp_web
 
 # 4MB is the default inside aiohttp
 MAX_MSG_SIZE = int(os.environ.get("WSLINK_MAX_MSG_SIZE", 4194304))
-HEART_BEAT = int(os.environ.get("WSLINK_HEART_BEAT", 30)) # 30 seconds
+HEART_BEAT = int(os.environ.get("WSLINK_HEART_BEAT", 30))  # 30 seconds
 
 
 async def _on_startup(app):
-    # Emit an expected log message so launcher.py knows we've started up.
-    print("wslink: Starting factory")
-    # We've seen some issues with stdout buffering - be conservative.
-    sys.stdout.flush()
+    STARTUP_MSG = os.environ.get("WSLINK_READY_MSG", "wslink: Starting factory")
+    if STARTUP_MSG:
+        # Emit an expected log message so launcher.py knows we've started up.
+        print(STARTUP_MSG)
+        # We've seen some issues with stdout buffering - be conservative.
+        sys.stdout.flush()
 
     # Also schedule server shutdown in case no clients connect within timeout
     _schedule_shutdown(app)
@@ -35,9 +38,7 @@ async def _on_startup(app):
 def _schedule_shutdown(app):
     timeout = app["state"]["server_config"]["timeout"]
     app["state"]["shutdown_task"] = (
-        schedule_coroutine(timeout, _stop_server, app)
-        if timeout > 0
-        else None
+        schedule_coroutine(timeout, _stop_server, app) if timeout > 0 else None
     )
 
 
@@ -48,6 +49,10 @@ async def _root_handler(request):
 
 
 async def _stop_server(app):
+    if hasattr(app, "stop"):
+        await app.stop()
+        return
+
     # Disconnecting any connected clients of handler(s)
     for route, handler in app["state"]["server_config"]["ws"].items():
         await handler.disconnectClients()
@@ -97,6 +102,7 @@ class AiohttpWslinkServer(object):
         port = int(self.config["port"])
         timeout = int(self.config["timeout"])
         handle_signals = self.config["handle_signals"]
+        ssl_context = self.config.get("ssl", None)
 
         runner = aiohttp_web.AppRunner(app, handle_signals=handle_signals)
         loop = asyncio.get_event_loop()
@@ -108,7 +114,7 @@ class AiohttpWslinkServer(object):
         logging.info("awaiting runner setup")
         await runner.setup()
 
-        my_site = aiohttp_web.TCPSite(runner, host, port)
+        my_site = aiohttp_web.TCPSite(runner, host, port, ssl_context=ssl_context)
 
         logging.info("awaiting site startup")
         await my_site.start()
@@ -122,8 +128,49 @@ class AiohttpWslinkServer(object):
     async def stop(self):
         await _stop_server(self.app)
 
+    @property
+    def completion(self):
+        return self.app["state"]["running"]
+
+
+class AiohttpWslinkClient(object):
+    def __init__(self, config):
+        self._config = config
+        self._url = config.get("reverse_url")
+        self._server_protocol = config.get("ws_protocol")
+        self._ws_handler = WslinkHandler(self._server_protocol, self)
+
+    def get_app(self):
+        return self
+
+    def get_config(self):
+        return self._config
+
+    def get_port(self):
+        return 0
+
+    async def start(self, port_callback=None):
+        if port_callback is not None:
+            port_callback(self.get_port())
+
+        await self._ws_handler.reverse_connect_to(self._url)
+
+    async def stop(self):
+        client_id = self._ws_handler.reverse_connection_client_id
+        ws = self._ws_handler.connections[client_id]
+        await ws.close()
+
 
 def create_webserver(server_config):
+    if "logging_level" in server_config and server_config["logging_level"]:
+        logging.basicConfig(level=server_config["logging_level"])
+
+    # Shortcut for reverse connection
+    if "reverse_url" in server_config:
+        server = AiohttpWslinkClient(server_config)
+        return server
+
+    # Normal web server
     web_app = aiohttp_web.Application()
 
     if "ws" in server_config:
@@ -152,9 +199,6 @@ def create_webserver(server_config):
         web_app.router.add_route("GET", "/", _root_handler)
         web_app.add_routes(routes)
 
-    if "logging_level" in server_config and server_config["logging_level"]:
-        logging.basicConfig(level=server_config["logging_level"])
-
     web_app.on_startup.append(_on_startup)
 
     web_app["state"] = {}
@@ -225,7 +269,9 @@ class WslinkHandler(object):
         aiohttp_app = request.app
 
         client_id = str(uuid.uuid4()).replace("-", "")
-        current_ws = aiohttp_web.WebSocketResponse(max_msg_size=MAX_MSG_SIZE, heartbeat=HEART_BEAT)
+        current_ws = aiohttp_web.WebSocketResponse(
+            max_msg_size=MAX_MSG_SIZE, heartbeat=HEART_BEAT
+        )
         self.connections[client_id] = current_ws
 
         logging.info("client {0} connected".format(client_id))
@@ -254,6 +300,31 @@ class WslinkHandler(object):
 
         return current_ws
 
+    @property
+    def reverse_connection_client_id(self):
+        return "reverse_connection_client_id"
+
+    async def reverse_connect_to(self, url):
+        logging.debug("reverse_connect_to: running with url %s", url)
+        client_id = self.reverse_connection_client_id
+        async with aiohttp.ClientSession() as session:
+            logging.debug("reverse_connect_to: client session started")
+            async with session.ws_connect(url) as current_ws:
+                logging.debug("reverse_connect_to: ws started")
+                self.connections[client_id] = current_ws
+                logging.debug("reverse_connect_to: onConnect")
+                await self.onConnect(url, client_id)
+
+                async for msg in current_ws:
+                    if not current_ws.closed:
+                        await self.onMessage(msg, client_id)
+
+                logging.debug("reverse_connect_to: onClose")
+                await self.onClose(client_id)
+                del self.connections[client_id]
+
+        logging.debug("reverse_connect_to: exited")
+
     async def onConnect(self, request, client_id):
         if not self.serverProtocol:
             return
@@ -327,7 +398,7 @@ class WslinkHandler(object):
             payload = payload.decode("utf-8")
 
         rpc = json.loads(payload)
-        logging.debug("wslink incoming msg %s" % payload)
+        logging.debug("wslink incoming msg %s" % self.payloadWithSecretStripped(rpc))
         if "id" not in rpc:
             # should be a binary attachment header
             if rpc.get("method") == "wslink.binary.attachment":
@@ -404,6 +475,11 @@ class WslinkHandler(object):
                 results = func(*args, **kwargs)
                 if inspect.isawaitable(results):
                     results = await results
+
+                if self.connections[client_id].closed:
+                    # Connection was closed during RPC call.
+                    return
+
                 await self.sendWrappedMessage(
                     rpcid, results, method=methodName, client_id=client_id
                 )
@@ -438,6 +514,14 @@ class WslinkHandler(object):
             )
             return
 
+    def payloadWithSecretStripped(self, payload):
+        payload = copy.deepcopy(payload)
+        if "args" in payload:
+            for arg in payload["args"]:
+                if type(arg) is dict and "secret" in arg:
+                    arg["secret"] = "*****"
+        return payload
+
     async def validateToken(self, token, client_id):
         if not self.serverProtocol:
             return True
@@ -507,10 +591,11 @@ class WslinkHandler(object):
                 # https://github.com/aio-libs/aiohttp/issues/2934
                 async with self.attachment_atomic:
                     for ws in websockets:
-                        # Send binary header
-                        await ws.send_str(json_header)
-                        # Send binary message
-                        await ws.send_bytes(attachments[key])
+                        if ws is not None:
+                            # Send binary header
+                            await ws.send_str(json_header)
+                            # Send binary message
+                            await ws.send_bytes(attachments[key])
 
                 # decrement for key
                 pub.publishManager.unregisterAttachment(key)
@@ -540,7 +625,8 @@ class WslinkHandler(object):
             else [self.connections[c] for c in self.connections]
         )
         for ws in websockets:
-            await ws.send_str(encMsg)
+            if ws is not None:
+                await ws.send_str(encMsg)
 
     def publish(self, topic, data, client_id=None):
         client_list = [client_id] if client_id else [c_id for c_id in self.connections]
